﻿; ------------------------------------------------------------------------------------
; Job Queue for Multi - Threading (Lite Edition)
; Author: Alexander Aigner
; PB 4.3+ x86/x64 Ascii/(Unicode)
; ------------------------------------------------------------------------------------

; ------------------------------------------------------------------------------------
; Prototypes, Structures & Enumerations
; ------------------------------------------------------------------------------------

XIncludeFile "Common\Stack.pbi"


Prototype JobQueue_Job(*JobData)

Structure _JobQueue_Job
  *Call.JobQueue_Job ; Job - Procedure
  *Info._JobQueue_JobInfo ; Job RückgabeInfo
  Para.i ; Übergabeparameter
  RealTime.a ; Job benötigt Echtzeit - Priorität
EndStructure

Structure _JobQueue_Worker
  Id.i ; ThreadID
  Number.i ; Worker Number
  DoCheck.i ; Check - Signal
  Mutex.i ; Worker Mutex
EndStructure

Structure _JobQueue_WorkerHolder
  Entry._JobQueue_Worker[0]
EndStructure

Structure _JobQueue_Data
  *Thread._JobQueue_WorkerHolder ; Worker Threads
  Mutex.i ; Kontroll Mutex
  *Stack.tStack[3] ; Stacks für 3 Prioritätsklassen
  Stop.i ; Deinit - Flag (= ThreadCount: Kill)
  Stoped.i ; Deinit - AnswerFlag
  Pause.i ; Pause - Flag
  Resume.i ; Resume - Signal
  WorkerCount.i ; Anzahl Worker
  JobCount.i ; Ausstehende Jobs
  JobsinProgress.i ; Jobs in Ausführung
  StopQueue.i ; Keine weiteren Jobs können mehr hinzugefügt werden
EndStructure

Structure _JobQueue_WorkerInit
  *JData._JobQueue_Data
  WorkerNumber.i ;Thread Nummer
EndStructure

Enumeration 0 ; Priority - Klassen
  #Job_LowPriority ; Niedrige Priorität, wird zuletzt bearbeitet
  #Job_NormalPriority ; Normale Prioriät, vor Low, hinter High
  #Job_HighPriority ; Hohe Priorität, wird zuerst behandelt
  #Job_HighestPriority ; Höchste Priorität: Wird an erste Stelle von High Priority gepusht
  #Job_RealTimePriority ; Höchste Priorität + Echtzeit!!!! Nur benutzen wenn Sie wissen was sie tun
EndEnumeration

; ------------------------------------------------------------------------------------
; Worker Internal
; ------------------------------------------------------------------------------------

; ist ein Check-Signal da?
Macro _JQ_CheckForCheckSignal()
  If *Worker\DoCheck
    LockMutex(*Worker\Mutex)
    *Worker\DoCheck = #False
    UnlockMutex(*Worker\Mutex)
    Break 2
  EndIf
EndMacro

; Sendet ein Check-Signal an einen Worker
Macro _JQ_SendCheckSignalToWorker(_JobQueue, _Worker)
  LockMutex(_JobQueue\Thread\Entry[_Worker]\Mutex)
  _JobQueue\Thread\Entry[_Worker]\DoCheck = #True
  UnlockMutex(_JobQueue\Thread\Entry[_Worker]\Mutex)
EndMacro

; Sendet ein Check-Signal an alle Worker
Macro _JQ_SendCheckSignal(_JobQueue)
  Protected cWorker
  For cWorker = 1 To _JobQueue\WorkerCount
    _JQ_SendCheckSignalToWorker(_JobQueue, cWorker)
  Next
EndMacro

; Worker Thread: Arbeitet die Jobs ab
; Der Worker bestimmt wie schnell die gesamte Queue ist, je weniger Overhead desto besser

Procedure JQ_Worker(*Init._JobQueue_WorkerInit)
  Protected oldPriority
  Protected *Job._JobQueue_Job, Priority
  Protected *JobQueue._JobQueue_Data = *Init\JData
  Protected *Worker._JobQueue_Worker = *JobQueue\Thread\Entry[*Init\WorkerNumber]
  Protected Mutex = *JobQueue\Mutex
  
  FreeMemory(*Init) ; Brauchen wir nicht mehr
  
  With *JobQueue
    Repeat ; Und los gehts
      If \Stop ; JobQueue entfernen
        LockMutex(Mutex)
        \Stoped + 1 ; Ein Worker weniger
        UnlockMutex(Mutex)
        ProcedureReturn #True ; beendet sich selbst
      ElseIf \Pause ; Workers Pausieren
        WaitSemaphore(\Resume) ; Warten auf Resume Signal
      EndIf
      
      ; Hier könnte bei zu vielen Threads das Reader-Writer Problem zu Verzögerungen
      ; durch zu viele Mutex Anfragen führen, dadurch das \JobsinProgress abgezogen
      ; wird minimiert sich das aber.
      
      If \JobCount-\JobsinProgress ; Ausführen wenn keine Pause ist und Jobs da sind
        LockMutex(Mutex) ; Mutex sperren
        For Priority = #Job_HighPriority To #Job_LowPriority Step -1 ; Höchste zu niedrigste Priorität
          While Stack_Count(\Stack[Priority]) ; Jobs abarbeiten
            _JQ_CheckForCheckSignal() ; Worker soll Flags & seine Messages checken
            
            \JobsinProgress + 1 ; möglichs früh setzen um das Reader-Writer Problem klein zu halten
            
            *Job = Stack_Pop(\Stack[Priority])
            
            UnlockMutex(Mutex) ; während des Calls muss die Mutex nicht gelockt sein
            
            If *Job\RealTime ; ECHTZEIT Priorität, ACHTUNG !!!
              oldPriority = ThreadPriority(*Worker\Id, 32)
            EndIf
            
            *Job\Call(*Job\Para) ; Job ausführen
            
            If *Job\RealTime
              ThreadPriority(*Worker\Id, oldPriority)
            EndIf
            
            LockMutex(Mutex) ; Mutex wieder sperren
            
            \JobCount-1; Einer mehr erledigt
            \JobsinProgress-1
            
            FreeMemory(*Job) ; Job Memory freigeben
            
          Wend
        Next
        UnlockMutex(Mutex) ; Mutex freigeben
      Else
        Delay(1) ; CPU schonen
      EndIf
      
    ForEver
  EndWith
EndProcedure

; ------------------------------------------------------------------------------------
; Public Functions
; ------------------------------------------------------------------------------------

; Initialisiert eine neue Job Queue, Worker = Anzahl Threads
Procedure JQ_Init(Worker = 1)
  Protected i, Mutex = CreateMutex(), *JobQueue._JobQueue_Data = AllocateMemory(SizeOf(_JobQueue_Data))
  Protected *Thread._JobQueue_WorkerInit
  LockMutex(Mutex) ; Mutex sperren
  If Worker< = 0
    Worker = 1
  EndIf
  With *JobQueue
    ; Stacks Initialisieren
    \Stack[#Job_LowPriority] = Stack_New()
    \Stack[#Job_NormalPriority] = Stack_New()
    \Stack[#Job_HighPriority] = Stack_New()
    \Resume = CreateSemaphore()
    \Mutex = Mutex
    \WorkerCount = Worker
    \Thread = AllocateMemory((Worker + 1)*SizeOf(_JobQueue_Worker))
    For i = 1 To Worker ; Threads starten
      *Thread = AllocateMemory(SizeOf(_JobQueue_WorkerInit))
      *Thread\WorkerNumber = i
      *Thread\JData = *JobQueue
      \Thread\Entry[i]\Number = i
      \Thread\Entry[i]\Mutex = CreateMutex()
      \Thread\Entry[i]\Id = CreateThread(@JQ_Worker(), *Thread) ; Worker starten
    Next
  EndWith
  UnlockMutex(Mutex) ; Mutex freigeben
  ProcedureReturn *JobQueue
EndProcedure

; Gibt eine Job Queue Frei
Procedure JQ_DeInit(*JobQueue._JobQueue_Data)
  Protected Priority, Worker, cTime
  With *JobQueue
    \Stop = #True ; Deinit - Flag setzen
    
    If \Pause
      \Pause = #False
      For Worker = 1 To \WorkerCount
        SignalSemaphore(\Resume)
      Next
    EndIf
    
    _JQ_SendCheckSignal(*JobQueue) ; Worker sollen Flags prüfen
    
    While \Stoped<>\WorkerCount
      Delay(1)
    Wend ; Warten bis alle Threads beendet haben
    
    For Worker = 1 To \WorkerCount
      LockMutex(*JobQueue\Thread\Entry[Worker]\Mutex)
      UnlockMutex(*JobQueue\Thread\Entry[Worker]\Mutex)
      FreeMutex(*JobQueue\Thread\Entry[Worker]\Mutex)
    Next
    For Priority = #Job_HighPriority To #Job_LowPriority Step -1 ; Stacks freigeben
      While Stack_Count(\Stack[Priority])
        FreeMemory(Stack_Pop(\Stack[Priority]))
      Wend
      Stack_Free(\Stack[Priority])
    Next
    
    LockMutex(\Mutex)
    UnlockMutex(\Mutex)
    
    ; Alles andere freigeben
    FreeSemaphore(\Resume)
    FreeMemory(\Thread)
    FreeMutex(\Mutex)
    FreeMemory(*JobQueue)
    ProcedureReturn #True
  EndWith
EndProcedure

; Pausiert die Jobverarbeitung
Procedure JQ_Pause(*JobQueue._JobQueue_Data)
  With *JobQueue
    If Not \Pause
      \Pause = #True
      _JQ_SendCheckSignal(*JobQueue)
      ProcedureReturn #True
    EndIf
  EndWith
EndProcedure

; Setz die Jobverarbeitung fort
Procedure JQ_Resume(*JobQueue._JobQueue_Data)
  Protected Workers, i
  With *JobQueue
    If \Pause
      \Pause = #False
      For i = 1 To \WorkerCount
        SignalSemaphore(\Resume)
      Next
      ProcedureReturn #True
    EndIf
  EndWith
EndProcedure

; Fügt einen Job hinzu (Rückgabewert: JobID für JobQueue_JobFinished()/JobQueue_JobReturnValue())
Procedure JQ_Add(*JobQueue._JobQueue_Data, *JobProcedure.JobQueue_Job, Parameter.i, Priority = #Job_NormalPriority)
  Protected *Job._JobQueue_Job = AllocateMemory(SizeOf(_JobQueue_Job))
  With *JobQueue
    ; Job erstellen
    *Job\Call = *JobProcedure
    *Job\Para = Parameter
    ;*Job\Info = 0
    
    LockMutex(\Mutex) ; Mutex sperren
    If \StopQueue
      FreeMemory(*Job)
      UnlockMutex(\Mutex)
      ProcedureReturn #False
    EndIf
    
    \JobCount + 1
    
    ; Job auf den Stack pushen
    If Priority = #Job_HighestPriority Or Priority = #Job_RealTimePriority
      If Priority = #Job_RealTimePriority
        *Job\RealTime = #True ; RealTime Priorität. ACHTUNG!!!!
      EndIf
      Stack_Push(\Stack[#Job_HighPriority], *Job) ; HighPriority, Erste Stelle
      _JQ_SendCheckSignal(*JobQueue) ; Sicher gleich verarbeiten
    Else
      Stack_PushF(\Stack[Priority], *Job)
    EndIf
    UnlockMutex(\Mutex); Mutex freigeben
    
    ; ProcedureReturn *Job
  EndWith
EndProcedure

; gibt die Anzahl der ausstehenden Jobs an
Procedure JQ_Count(*JobQueue._JobQueue_Data)
  With *JobQueue
    ProcedureReturn \JobCount
  EndWith
EndProcedure

Procedure JQ_WaitUntilFinish(*JobQueue._JobQueue_Data, StopQueue = #True)
  
  With *JobQueue
    LockMutex(\Mutex)
    \StopQueue = StopQueue
    UnlockMutex(\Mutex)
    While JQ_Count(*JobQueue)
      Delay(1)  
    Wend
  EndWith
EndProcedure
; IDE Options = PureBasic 4.51 (Windows - x64)
; CursorPosition = 126
; FirstLine = 105
; Folding = --
; EnableThread
; EnableXP
; Executable = E:\Dokumente und Einstellungen\Alexander\Desktop\demo.exe
; CPU = 1
; DisableDebugger
; EnableCompileCount = 864
; EnableBuildCount = 1
; EnableExeConstant